$ifndef _DTMF_OUT_INC_
$define _DTMF_OUT_INC_
' -----------------------------------------------------------------------------------------------------------
'| This source code is provided 'as is' and any express or implied warranties, including,                    |
'| but not limited to, the implied warranties of merchantability and fitness for a particular                |
'| purpose are disclaimed.                                                                                   |
'|                                                                                                           |
'| In no event shall the author, Crownhill or contributers be liable for any direct, indirect, incidental,   |
'| special, exemplary, or consequential damages (including, but not limited to,  procurement of substitute   |
'| goods or services; loss of use, data, or profits;  or business interruption) however caused and on any    |
'| theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise)   |
'| arising in any way out of the use of this source code, even if advised of the possibility of such damage. |
'|                                                                                                           |
'| Any views and conclusions contained in the source code and any documentation are those of the author's    |
'| and should not be interpreted as representing official policies,                                          |
'| either expressed or implied, of Crownhill Associates Ltd.                                                 |
' -----------------------------------------------------------------------------------------------------------
'
' 24-bit DDS algorithm for DTMF dual simultaneous sinewaves on one PWM channel (CCP1)
' The routines produce quite pure sine waves
'
' Written by Les Johnson (21-04-14)
' For use with the Proton BASIC compiler. Only for use with an 18F device.
'
'         |     2.7K
' PORTC.2 |----/\/\/\-----o-----> Output
'         |               |
'         |             __|__
'         |             _____ 100nF
'         |               |
'         |               |
'         |             -----
'                        ---  Gnd
'                         -
' DDS values:
' Calculated by:
'
' fDivisor = (Device MIPS * 1000000) / 128
' DDS value = (Frequency(in Hz) * 64) * (65536 / fDivisor)
'
' Note that the values below have been calculated for an oscillator of 64MHz (16 MIPS)
'
    $define cDTMF_FreqRow0 23387                ' 697 Hz DTMF rows
    $define cDTMF_FreqRow1 25837                ' 770 Hz
    $define cDTMF_FreqRow2 28588                ' 852 Hz
    $define cDTMF_FreqRow3 31575                ' 941 Hz

    $define cDTMF_FreqCol0 40567                ' 1209 Hz DTMF columns
    $define cDTMF_FreqCol1 44829                ' 1336 Hz
    $define cDTMF_FreqCol2 49560                ' 1477 Hz
    $define cDTMF_FreqCol3 54794                ' 1633 Hz
'
' Create variables for use with the DTMF routines
'
    Dim DTMF_dAccum1 As Dword System            ' 32-bit accumulator for the row sinewave
    Dim DTMF_dAccum2 As Dword System            ' 32-bit accumulator for the column sinewave
    Dim DTMF_bPWM As Byte System                ' PWM accumulative value
    Dim DTMF_bTemp As Byte System               ' Temporary storage
    Dim DTMF_dDuration As Dword System          ' Main duration counter

    Dim DTMF_wRowValue As Word                  ' Holds the row value for decoding keypresses
    Dim DTMF_wColValue As Word                  ' Holds the column value for decoding keypresses

    Dim DTMF_bKey As Byte                       ' Holds the key value
    Dim DTMF_wSave As Word                      ' Storage for saving/restoreing FSR0L\H or TBLPTRL\H
    Dim DTMF_wFSR0 As FSR0L.Word                ' Create a 16-bit SFR from FSR0L\H
    Dim DTMF_wTblptr As TBLPTRL.Word            ' Create a 16-bit SFR from TBLPTRL\H
'
' Row frequency sine wave
'
    Dim cDTMF_SineLow As Code = 39, 42, 46, 50, 53, 57, 60, 63,_
                                66, 68, 71, 72, 74, 75, 76, 77,_
                                77, 77, 76, 75, 74, 72, 71, 68,_
                                66, 63, 60, 57, 53, 50, 46, 42,_
                                39, 35, 31, 27, 24, 20, 17, 14,_
                                11, 09, 06, 05, 03, 02, 01, 00,_
                                00, 00, 01, 02, 03, 05, 06, 09,_
                                11, 14, 17, 20, 24, 27, 31, 35
'
' Column frequency sine wave
' Same sine wave as above, but 28% amplitude increase
'
    Dim cDTMF_SineHigh As Code = 50, 54, 59, 64, 68, 73, 77, 81,_
                                 85, 88, 91, 93, 95, 97, 98, 99,_
                                 99, 99, 98, 97, 95, 93, 91, 88,_
                                 85, 81, 77, 73, 68, 64, 59, 54,_
                                 50, 45, 40, 35, 31, 26, 22, 18,_
                                 14, 11, 08, 06, 04, 02, 01, 00,_
                                 00, 00, 01, 02, 04, 06, 08, 11,_
                                 14, 18, 22, 26, 31, 35, 40, 45

'-----------------------------------------------------------------------------------------
    GoTo _DTMF_Main_                                                ' Jump over the subroutines
'-----------------------------------------------------------------------------------------
' Initialise the program
' Input     : None
' Output    : None
' Notes     : Configures and enables the PWM peripheral
'
$define DTMF_Init() GoSub _DTMF_Init

_DTMF_Init:
    T2CON = %00000100
    PR2 = 127
    CCPR1L = 0
    CCP1CON = %00001100                                             ' Enable the PWM peripheral
    Output PORTC.2                                                  ' Make the PWM peripheral's pin an output
    Return

'-----------------------------------------------------------------------------------------
' Generate a two frequency DTMF signal from the CCP1 peripheral pin
' Input     : pRowFreq is the DDS value for the row frequency
'           : pColFreq is the DDS value For the column frequency
'           : pDuration is the length of time the tone will be sent (in ms)
' Output    : None
' Notes     : None
'
$define DTMF_Gen(pRowFreq, pColFreq, pDuration) '
    DTMF_wRowValue = pRowFreq                   '
    DTMF_wColValue = pColFreq                   '
    DTMF_dDuration = pDuration                  '
    GoSub _DTMF_Generate

_DTMF_Generate:
    DTMF_dDuration = DTMF_dDuration * 124                           ' Calculate the duration in approx milliseconds
    DTMF_dAccum1 = 0                                                ' \ Reset the accumulators
    DTMF_dAccum2 = 0                                                ' /
    While                                                           ' Create an infinite loop
        While PIR1bits_TMR2IF <> 1: Wend                            ' \ Synchronise to the start of the PWM cycle
        PIR1bits_TMR2IF = 0                                         ' /
        '
        ' Calculate the row sine wave
        '
        DTMF_dAccum1 = DTMF_dAccum1 + DTMF_wRowValue                ' Accumulation of the row frequency
        DTMF_bPWM = CRead8 cDTMF_SineLow[DTMF_dAccum1.Byte2 & $3F]  ' Binary Divide output by 65536 and keep the first 6 bits
        '
        ' Calculate the column sine wave
        '
        DTMF_dAccum2 = DTMF_dAccum2 + DTMF_wColValue                ' Accumulation of the column frequency
        DTMF_bTemp = CRead8 cDTMF_SineHigh[DTMF_dAccum2.Byte2 & $3F]' Binary Divide output by 65536 and keep the first 6 bits

        DTMF_bPWM = DTMF_bPWM + DTMF_bTemp                          ' Add the 2 frequencies together
        DTMF_bPWM = DTMF_bPWM >> 1                                  ' Scale 0-200 back to 0-100 for PWM
        CCPR1L = DTMF_bPWM                                          ' Transfer the signal to the PWM peripheral
        Dec DTMF_dDuration                                          ' Decrement the duration counter
        If DTMF_dDuration = 0 Then                                  ' Has the duration reached 0?
            Return                                                  ' Yes. So exit the subroutine
        EndIf
    Wend
'-----------------------------------------------------------------------------------------
' Generate a two-tone signal for a particular key value
' Input     : pKey is the ASCII key value (1,2,3,4,5,6,7,8,9,A,B,C,D,*,#)
'           : pDuration is the duration of the tone (in approx ms)
' Output    : None
' Notes     : Uses a crude, but effective set of comparisons for decoding
'
$define DTMF_KeyOut(pKey, pDuration) '
    DTMF_bKey = pKey                 '
    DTMF_dDuration = pDuration       '
    GoSub _DTMF_Out

_DTMF_Out:
    If DTMF_bKey = " " Then                             ' Is the key a space?
        DelayMS DTMF_dDuration                          ' Yes. So create a delay
    Else                                                ' Otherwise...
        Select DTMF_bKey
            Case "1"                                    ' Is the key "1"?
                DTMF_wRowValue = cDTMF_FreqRow0         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol0         ' Set the column frequency
            Case "2"                                    ' Is the key "2"?
                DTMF_wRowValue = cDTMF_FreqRow0         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol1         ' Set the column frequency
            Case "3"                                    ' Is the key "3"?
                DTMF_wRowValue = cDTMF_FreqRow0         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol2         ' Set the column frequency
            Case "A"                                    ' Is the key "A"?
                DTMF_wRowValue = cDTMF_FreqRow0         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol3         ' Set the column frequency
            Case "4"                                    ' Is the key "4"?
                DTMF_wRowValue = cDTMF_FreqRow1         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol0         ' Set the column frequency
            Case "5"                                    ' Is the key "5"?
                DTMF_wRowValue = cDTMF_FreqRow1         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol1         ' Set the column frequency
            Case "6"                                    ' Is the key "6"?
                DTMF_wRowValue = cDTMF_FreqRow1         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol2         ' Set the column frequency
            Case "B"                                    ' Is the key "B"?
                DTMF_wRowValue = cDTMF_FreqRow1         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol3         ' Set the column frequency
            Case "7"                                    ' Is the key "7"?
                DTMF_wRowValue = cDTMF_FreqRow2         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol0         ' Set the column frequency
            Case "8"                                    ' Is the key "8"?
                DTMF_wRowValue = cDTMF_FreqRow2         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol1         ' Set the column frequency
            Case "9"                                    ' Is the key "9"?
                DTMF_wRowValue = cDTMF_FreqRow2         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol2         ' Set the column frequency
            Case "C"                                    ' Is the key "C"?
                DTMF_wRowValue = cDTMF_FreqRow2         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol3         ' Set the column frequency
            Case "*"                                    ' Is the key "*"?
                DTMF_wRowValue = cDTMF_FreqRow3         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol0         ' Set the column frequency
            Case "0"                                    ' Is the key "0"?
                DTMF_wRowValue = cDTMF_FreqRow3         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol1         ' Set the column frequency
            Case "#"                                    ' Is the key "#"?
                DTMF_wRowValue = cDTMF_FreqRow3         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol2         ' Set the column frequency
            Case "D"                                    ' Is the key "D"?
                DTMF_wRowValue = cDTMF_FreqRow3         ' Yes. So set the row frequency
                DTMF_wColValue = cDTMF_FreqCol3         ' Set the column frequency
        EndSelect
        GoSub _DTMF_Generate                            ' Generate the tones
    EndIf
    Return

'-----------------------------------------------------------------------------------------
' Generate DTMF signals from a RAM string holding the key values
' Input     : pStringIn holds the string containing the key values (1,2,3,4,5,6,7,8,9,A,B,C,D,*,#)
' Output    : None
' Notes     : Duration for On time is set for approx 200ms
'           : Duration for Off time is set to 100ms
'
$define DTMF_String(pStringIn)        '
    DTMF_wFSR0 = AddressOf(pStringIn) '
    GoSub _DTMF_String

_DTMF_String:
    While                                   ' Create a loop
        DTMF_dDuration = 200                ' Set the duration to approx 200ms
        If INDF0 = 0 Then Break             ' Exit the loop when the null is found
        DTMF_bKey = POSTINC0                ' Transfer the key value to DTMF_bKey
        DTMF_wSave = DTMF_wFSR0             ' Save FSR0L\H in case they are altered
        GoSub _DTMF_Out                     ' Generate the relevant tones
        DTMF_wFSR0 = DTMF_wSave             ' Restore FSR0L\H
        DelayMS 100                          ' Set the off time delay to 100ms
    Wend                                    ' Close the loop
    Return
'-----------------------------------------------------------------------------------------
' Generate DTMF signals from a code memory string holding the key values
' Input     : pCodeIn holds the code memory string containing the key values (1,2,3,4,5,6,7,8,9,A,B,C,D,*,#)
' Output    : None
' Notes     : Duration for On time is set for approx 200ms
'           : Duration for Off time is set to 100ms
'
$define DTMF_Code(pCodeIn)            '
    DTMF_wTblptr = AddressOf(pCodeIn) '
    GoSub _DTMF_Code

_DTMF_Code:
    While                                   ' Create a loop
        DTMF_dDuration = 200                ' Set the duration to approx 200ms
        Tblrd*+                             ' Read a byte from code memory into TABLAT
        If TABLAT = 0 Then Break            ' Exit the loop when the null is found
        DTMF_bKey = TABLAT                  ' Transfer the key value to DTMF_bKey
        DTMF_wSave = DTMF_wTblptr           ' Save TBLPTRL\H
        GoSub _DTMF_Out                     ' Generate the relevant tones
        DTMF_wTblptr = DTMF_wSave           ' Restore TBLPTRL\H
        DelayMS 100                          ' Set the off time delay to 100ms
    Wend                                    ' Close the loop
    Return

'-----------------------------------------------------------------------------------------
_DTMF_Main_:
    DTMF_Init()                             ' Initialise the PWM peripheral

$endif  ' _DTMF_OUT_INC_


